package dev.fyloz.colorrecipesexplorer.config.initializers

import dev.fyloz.colorrecipesexplorer.config.annotations.RequireDatabase
import dev.fyloz.colorrecipesexplorer.config.properties.MaterialTypeProperties
import dev.fyloz.colorrecipesexplorer.model.MaterialType
import dev.fyloz.colorrecipesexplorer.model.materialType
import dev.fyloz.colorrecipesexplorer.service.MaterialTypeService
import mu.KotlinLogging
import org.springframework.context.annotation.Configuration

@Configuration
@RequireDatabase
class MaterialTypeInitializer(
    private val materialTypeService: MaterialTypeService,
    private val materialTypeProperties: MaterialTypeProperties
) : AbstractInitializer() {
    private val logger = KotlinLogging.logger {}

    override fun initialize() {
        logger.debug("Executing material type initializer...")
        ensureSystemMaterialTypesExists()
        logger.debug("System material types are up to date!")
    }

    private fun ensureSystemMaterialTypesExists() {
        val systemTypes = materialTypeProperties.systemTypes.map { it.toMaterialType() }
        val oldSystemTypes = materialTypeService.getAllSystemTypes().toMutableSet()

        fun saveOrUpdateSystemType(type: MaterialType) {
            if (materialTypeService.existsByName(type.name)) {
                with(materialTypeService.getByName(type.name)) {
                    if (!this.systemType) {
                        logger.info("Material type '${type.name}' already exists and will be flagged as a system type")
                        materialTypeService.update(this.copy(systemType = true))
                    } else {
                        logger.debug("System material type '${type.name}' already exists")
                    }
                }
            } else {
                logger.info("System material type '${type.name}' will be created")
                materialTypeService.save(type)
            }
        }

        // Save new system types
        systemTypes.forEach {
            saveOrUpdateSystemType(it)
            oldSystemTypes.removeIf { c -> c.name == it.name }
        }

        // Remove old system types
        oldSystemTypes.forEach {
            logger.info("Material type '${it.name}' is not a system type anymore")
            materialTypeService.updateSystemType(it.copy(systemType = false))
        }
    }
}