package dev.fyloz.colorrecipesexplorer.service.files

import dev.fyloz.colorrecipesexplorer.JavaFile
import dev.fyloz.colorrecipesexplorer.utils.File
import dev.fyloz.colorrecipesexplorer.utils.FilePath
import dev.fyloz.memorycache.MemoryCache
import mu.KotlinLogging
import org.springframework.stereotype.Component

interface FileCache {
    /** Checks if the cache contains the given [path]. */
    operator fun contains(path: FilePath): Boolean

    /** Gets the cached file system item at the given [path]. */
    operator fun get(path: FilePath): CachedFileSystemItem?

    /** Gets the cached directory at the given [path]. */
    fun getDirectory(path: FilePath): CachedDirectory?

    /** Gets the cached file at the given [path]. */
    fun getFile(path: FilePath): CachedFile?

    /** Checks if the cached file system item at the given [path] exists. */
    fun exists(path: FilePath): Boolean

    /** Checks if the cached directory at the given [path] exists. */
    fun directoryExists(path: FilePath): Boolean

    /** Checks if the cached file at the given [path] exists. */
    fun fileExists(path: FilePath): Boolean

    /** Sets the file system item at the given [path] as existing or not. Loads the item in the cache if not already present. */
    fun setExists(path: FilePath, exists: Boolean = true)

    /** Loads the file system item at the given [path] into the cache. */
    fun load(path: FilePath)

    /** Adds the file system item at the given [itemPath] to the cached directory at the given [directoryPath]. */
    fun addItemToDirectory(directoryPath: FilePath, itemPath: FilePath)

    /** Removes the file system item at the given [itemPath] from the cached directory at the given [directoryPath]. */
    fun removeItemFromDirectory(directoryPath: FilePath, itemPath: FilePath)
}

@Component
class DefaultFileCache(private val cache: MemoryCache<String, CachedFileSystemItem>) : FileCache {
    private val logger = KotlinLogging.logger {}

    override operator fun contains(path: FilePath) =
        path.value in cache

    override operator fun get(path: FilePath) =
        cache[path.value]

    private operator fun set(path: FilePath, item: CachedFileSystemItem) {
        cache[path.value] = item
    }

    override fun getDirectory(path: FilePath) =
        if (directoryExists(path)) {
            this[path] as CachedDirectory
        } else {
            null
        }

    override fun getFile(path: FilePath) =
        if (fileExists(path)) {
            this[path] as CachedFile
        } else {
            null
        }

    override fun exists(path: FilePath) =
        path in this && this[path]!!.exists

    override fun directoryExists(path: FilePath) =
        exists(path) && this[path] is CachedDirectory

    override fun fileExists(path: FilePath) =
        exists(path) && this[path] is CachedFile

    override fun setExists(path: FilePath, exists: Boolean) {
        if (path !in this) {
            load(path)
        }

        this[path] = this[path]!!.clone(exists = exists)
        logger.debug("Updated FileCache state: ${path.value} exists -> $exists")
    }

    override fun load(path: FilePath) =
        with(JavaFile(path.value).toFileSystemItem()) {
            this@DefaultFileCache[path] = this

            logger.debug("Loaded file at ${path.value} into FileCache")
        }

    override fun addItemToDirectory(directoryPath: FilePath, itemPath: FilePath) {
        val directory = prepareDirectory(directoryPath) ?: return

        val updatedContent = setOf(
            *directory.content.toTypedArray(),
            JavaFile(itemPath.value).toFileSystemItem()
        )

        this[directoryPath] = directory.copy(content = updatedContent)
        logger.debug("Added child ${itemPath.value} to ${directoryPath.value} in FileCache")
    }

    override fun removeItemFromDirectory(directoryPath: FilePath, itemPath: FilePath) {
        val directory = prepareDirectory(directoryPath) ?: return

        val updatedContent = directory.content
            .filter { it.path.value != itemPath.value }
            .toSet()

        this[directoryPath] = directory.copy(content = updatedContent)
        logger.debug("Removed child ${itemPath.value} from ${directoryPath.value} in FileCache")
    }

    private fun prepareDirectory(path: FilePath): CachedDirectory? {
        if (!directoryExists(path)) {
            logger.warn("Cannot add child to ${path.value} because it is not in the cache")
            return null
        }

        val directory = getDirectory(path)
        if (directory == null) {
            logger.warn("Cannot add child to ${path.value} because it is not a directory")
            return null
        }

        return directory
    }
}

interface CachedFileSystemItem {
    val name: String
    val path: FilePath
    val exists: Boolean

    fun clone(exists: Boolean): CachedFileSystemItem
}

data class CachedFile(
    override val name: String,
    override val path: FilePath,
    override val exists: Boolean
) : CachedFileSystemItem {
    constructor(file: File) : this(file.name, file.toFilePath(), file.exists() && file.isFile)

    override fun clone(exists: Boolean) =
        this.copy(exists = exists)
}

data class CachedDirectory(
    override val name: String,
    override val path: FilePath,
    override val exists: Boolean,
    val content: Set<CachedFileSystemItem> = setOf()
) : CachedFileSystemItem {
    constructor(file: File) : this(file.name, file.toFilePath(), file.exists() && file.isDirectory, file.fetchContent())

    val contentFiles: Collection<CachedFile>
        get() = content.filterIsInstance<CachedFile>()

    override fun clone(exists: Boolean) =
        this.copy(exists = exists)

    companion object {
        private fun File.fetchContent() =
            (this.file.listFiles() ?: arrayOf<JavaFile>())
                .filterNotNull()
                .map { it.toFileSystemItem() }
                .toSet()
    }
}

fun JavaFile.toFileSystemItem() =
    if (this.isDirectory) {
        CachedDirectory(File(this))
    } else {
        CachedFile(File(this))
    }